/*****************************************************************

Copyright (c) 2001 John Firebaugh <jfirebaugh@kde.org>
Copyright (c) 2002 Siegfried Nijssen <snijssen@liacs.nl>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#include <qlayout.h>
#include <qwidgetstack.h>
#include <kdebug.h>
#include <karrowbutton.h>
//#include <kwindowlistmenu.h>
#include <kconfig.h>
#include <kapplication.h>
#include <kiconloader.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#define XK_LATIN1
#include <X11/keysymdef.h>
#include <X11/Xutil.h>

#include "menubarcontainer.h"
#include "menubarcontainer.moc"
#include <kapplication.h>
#include <kipc.h>
#include <dcopclient.h>
#include <qxembed.h>
#include <qmenubar.h>
#include <qtimer.h>
#include <qptrdict.h>
#include <qlabel.h>
#include <qtoolbutton.h>
#include <taskmanager.h>
#include <kwinmodule.h>
#include <kaboutkde.h>
#include <krun.h>

//#define WINDOWLISTBUTTON_SIZE 12

#define BUTTON_MIN_HEIGHT 18
#define BUTTON_HEIGHT 20
#define BUTTON_MAX_WIDTH 200
#define BUTTON_MIN_WIDTH 20

class MyTaskManager:public TaskManager {
   public:
     MyTaskManager( QObject *parent = 0, const char *name = 0 ):
       TaskManager( parent, name ) { }
     Task* findTask(WId w) { return TaskManager::findTask ( w ); }
};

MyTaskManager *MenuBarContainer::taskmanager = 0;
TaskList ctasklist;
extern KWinModule *kwin_module; // reuse from taskmanager

MenuBarContainer::MenuBarContainer( bool enableFrame, QWidget *parent, const char *name )
    : FittsLawFrame( parent, name )
    , DCOPObject ( QCString ( "MenubarApplet" ) )
    , direction( KPanelApplet::Up )
{
    currenttask = NULL;
    setFrameStyle( Box );
    setLineWidth ( 0 );

    layout = new QBoxLayout( this, QApplication::reverseLayout() ?
                                   QBoxLayout::RightToLeft :
                                   QBoxLayout::LeftToRight );
    layout->setAlignment ( Qt::AlignVCenter );
    layout->setMargin( 0 );
    applabel = new QLabel ( this );
    applabel->setSizePolicy ( QSizePolicy ( QSizePolicy::Fixed, QSizePolicy::Fixed ) );
    applabel->setIndent ( 10 );
    applabel->setText ( "<b>KDE</b>" );
    applabel->show ();
    layout->addWidget ( applabel );

    menubars = new QWidgetStack ( this, "menubars" );
    menubars->setBackgroundMode ( PaletteButton );
    layout->addWidget ( menubars );
    defaultMenubar = new QMenuBar ( NULL, "defaultMenubar" );
    editmenu = new QPopupMenu;
    editmenu->insertItem ( SmallIcon("editcopy"), "&Copy Ctrl+C", this, SLOT(copy()) );
    editmenu->insertItem ( SmallIcon("editpaste"), "&Paste Ctrl+V", this, SLOT(paste()) );
    defaultMenubar->insertItem ( "Edit", editmenu, 1 );
    windowmenu = new QPopupMenu;
    windowmenu->insertItem ( SmallIcon("window_new"), "Start New", this, SLOT(newWindow()), 0, 1 );
    windowmenu->insertSeparator ();
    otherwindowsmenu = new QPopupMenu;
    otherwindowsmenumanager = new TasksMenuManager ( otherwindowsmenu, currenttask, TasksMenuManager::OtherThanTask );
    windowmenu->insertItem ( "Other Windows", otherwindowsmenu );
    windowmenu->insertSeparator ();
    windowmenumanager = new TaskMenuManager ( windowmenu, currenttask );
    defaultMenubar->insertItem ( "Window", windowmenu, 2 );
    allwindowmenu = NULL;

    helpmenu = new QPopupMenu;
    helpmenu->insertItem ( BarIcon( "contents", KIcon::SizeSmall), "Manual", this, SLOT(helpCurrentWindow()), 0, 1 );
    helpmenu->insertItem ( SmallIcon("about_kde"), "About KDE", this, SLOT(helpAboutKDE()), 0, 2 );
    aboutkde = NULL;
    defaultMenubar->insertItem ( "Help", helpmenu, 3 );
    noWindowMenubar = 0L;

    menubars->addWidget ( defaultMenubar );
    menubars->setFixedHeight ( defaultMenubar->height () - 6 );
    // currently this high value (6) seems necessary to line up the label and the menu
    // in most styles.
    slotChangeStyle ();
    menubars->hide ();


    // the taskmanager is not used for menubar raising and lowering.
    // It is mainly used to provide a nice default Menubar.
    connect(kwin_module, SIGNAL(activeWindowChanged(WId)), SLOT(activeWindowChanged(WId)));
    connect(kwin_module, SIGNAL(taskRemoved(Task*)), SLOT(removeMenubar(Task*)));

    connect( kapp, SIGNAL(kdisplayStyleChanged()),
	     this, SLOT(slotChangeStyle()));

    configure();

    KIPC::sendMessageAll(KIPC::ToolbarStyleChanged);
}

MenuBarContainer::~MenuBarContainer()
{
    setObjId ( QCString ( "MenubarAppletDeleted" ) );
    KIPC::sendMessageAll(KIPC::ToolbarStyleChanged);
}

void MenuBarContainer::configure()
{
    KConfig c( "kmenubarrc", false, false );
    c.setGroup( "General" );
}

void MenuBarContainer::preferences()
{
//    kapp->startServiceByDesktopName( "kcmtaskbar", QStringList(),
//                                     0, 0, 0, "", true);
}

void MenuBarContainer::orientationChange( Orientation o )
{
    if( o == Horizontal ) {
	layout->setDirection( QApplication::reverseLayout() ?
	                      QBoxLayout::RightToLeft :
	                      QBoxLayout::LeftToRight );
    } else {
	layout->setDirection( QBoxLayout::TopToBottom );
    }
    layout->activate();
}

QSize MenuBarContainer::sizeHint( KPanelExtension::Position p, QSize maxSize) const
{
    return menubars->sizeHint ();
}

int MenuBarContainer::widthForHeight ( int height ) {
    return 200;
}


WId MenuBarContainer::addMenubar ( WId menuid, WId toplevelid ) {
    QXEmbed *newembed = new QXEmbed ( menubars );
    newembed->setBackgroundMode ( PaletteButton );
    menubars->addWidget ( newembed, toplevelid );
    newembed->embed ( menuid );
    newembed->show ();

    return winId ();
}

// toplevelid is the id of the window to which the menubar belongs.
// As the no window-menubar is typically the desktop's menubar, this
// may be confusing. However, also in that case there is such a window.
void MenuBarContainer::setNoWindowMenubar ( WId toplevelid ) {
    noWindowMenubar = toplevelid; // we use the id as the widget may not be known to the applet yet
}

void MenuBarContainer::removeMenubar ( Task *task ) {
    WId id = task->window ();
    QWidget *embed = menubars->widget ( id );
    if ( embed ) {
        menubars->removeWidget ( embed );
        delete embed;
    }
}

void MenuBarContainer::raiseDefaultMenubar () {
    if ( currenttask ) {
        QString name = currenttask->classClass ();
	name[0] = name[0].upper ();
	helpmenu->changeItem ( 1, name + " Manual" );
	windowmenu->changeItem ( 1, QString ( "Start New " ) + name );
	createAllWinMenu ( name );
	if ( otherwindowsmenumanager )
	    delete otherwindowsmenumanager;
	otherwindowsmenumanager = new TasksMenuManager ( otherwindowsmenu, currenttask, TasksMenuManager::OtherThanTask );
        if ( windowmenumanager )
	    delete windowmenumanager;
        windowmenumanager = new TaskMenuManager ( windowmenu, currenttask );
        defaultMenubar->setBackgroundMode( PaletteButton );
        menubars->raiseWidget ( defaultMenubar );
    }
}

void MenuBarContainer::copy () {
    sendkey ( 0, 37 );
    sendkey ( ControlMask, 54 );
}


void MenuBarContainer::sendkey ( long mask, long key ) {
    XEvent e;
    e.xkey.type = KeyPress;
    e.xkey.state = mask;
    e.xkey.keycode = key;
    e.xkey.display = qt_xdisplay ();
    e.xkey.window = currenttask->window ();
    e.xkey.root = qt_xrootwin ();
    e.xkey.subwindow = 0;
    e.xkey.same_screen = true;
    e.xkey.send_event = true;
    KWin::Info info = KWin::info ( currenttask->window () );
    e.xkey.x = info.geometry.width () / 2;
    e.xkey.y = info.geometry.height () / 2;
    e.xkey.x_root = info.geometry.x () + e.xkey.x;
    e.xkey.y_root = info.geometry.y () + e.xkey.y;
    XSendEvent ( qt_xdisplay (), currenttask->window (), true, KeyPressMask, &e );
}

void MenuBarContainer::paste () {
    sendkey ( 0, 37 );
    sendkey ( ControlMask, 55 );
}


MyTaskManager *MenuBarContainer::taskManager () {
    if ( !taskmanager )
    	taskmanager = new MyTaskManager ( this );
    return taskmanager;
}

void MenuBarContainer::activeWindowChanged ( WId w ) {
    currenttask = taskManager () ->findTask ( w );
    if ( currenttask ) {
	QString oname = currenttask->classClass ();
	QString name = "";
	name[0] = oname[0].upper ();
	for ( int i = 1; i < oname.length(); i++ )
	     if ( oname[i].isSpace () )
	         oname[i+1] = oname[i+1].upper ();
	     else
	         name += oname[i];

	// the label should be on 1 line.
        applabel->setText ( QString ( "<b>") + name + "</b>" );

	menubars->show ();
	if ( menubars->widget ( w ) ) {
	  menubars->raiseWidget ( w );
          menubars->repaint ();
          layout->activate ();
	}
	else
	  raiseDefaultMenubar ();
    }
    else {
        applabel->setText ( "<b>Desktop</b>" );
	if ( noWindowMenubar ) {
	    menubars->show ();
	    menubars->raiseWidget ( noWindowMenubar );
	    layout->activate ();
	}
	else
	    menubars->hide ();
    }
}

void MenuBarContainer::helpCurrentWindow () {
    QString name ( "man:" );
    name += currenttask->classClass ().lower ();
    new KRun ( name );
}

void MenuBarContainer::helpAboutKDE () {
    if ( !aboutkde )
        aboutkde = new KAboutKDE ( this, 0, false );
    aboutkde->show ();
}

void MenuBarContainer::newWindow () {
    KRun::runCommand ( currenttask->classClass ().lower () );
}

void MenuBarContainer::createAllWinMenu ( const QString &name ) {
    bool andere = false;
    if ( allwindowmenu ) {
        windowmenu->removeItem ( 2 );
	delete allwindowmenu;
    }
    allwindowmenu = new QPopupMenu;
    ctasklist = taskmanager->tasks ();
    QPtrListIterator<Task> it ( ctasklist );
    for ( ; it.current (); ++it ) {
        if ( Task::idMatch ( it.current ()->classClass (), currenttask->classClass () ) ) {
	    if ( it.current () != currenttask )
	        andere = true;
	    QPopupMenu *p2 = new QPopupMenu ( allwindowmenu );
	    connect ( new TaskMenuManager ( p2, it.current () ), SIGNAL(remove(int)), this, SLOT(removeTaskMenu(int)) );
	    allwindowmenu->insertItem ( it.current ()->visibleName (), p2, it.current ()->window () );
	}
    }
    QPopupMenu *o = new QPopupMenu ( allwindowmenu );
    new TasksMenuManager ( o, currenttask, TasksMenuManager::OtherThanTaskGroup );
    allwindowmenu->insertItem ( "Other Windows", o );
    allwindowmenu->insertSeparator ();
    new TasksMenuManager ( allwindowmenu, currenttask, TasksMenuManager::AllOfTaskGroup );
    if ( andere )
        windowmenu->insertItem ( QString ( "All " ) + name + " Windows", allwindowmenu, 2, 2 );
    else {
        delete allwindowmenu;
	allwindowmenu = NULL;
    }
}

void MenuBarContainer::removeTaskMenu ( int id ) {
    if ( allwindowmenu ) {
        allwindowmenu->removeItem ( id ); // not yet deleted, this will happen later
    }
}


void MenuBarContainer::slotChangeStyle () {
    defaultMenubar->setLineWidth( 0 );
    defaultMenubar->setFrameStyle ( Box );
    defaultMenubar->setBackgroundMode( PaletteBackground );

}

#define TMM_CLOSE_MENU 1200
#define TMM_KILL_MENU 1201
#define TMM_MAXIMIZE_MENU 1202
#define TMM_MINIMIZE_MENU 1203
#define TMM_RESTORE_MENU 1204
#define TMM_ACTIVATE_MENU 1205
#define TMM_SHADE_MENU 1206
#define TMM_ALWAYSONTOP_MENU 1207
#define TMM_DESKTOP_MENU 1208
#define add_item(id,text2,slot) \
    if ( !menu->text ( id ).isNull () && menu->indexOf ( id ) != -1 ) {\
        menu->removeItem ( id );\
    }\
    menu->insertItem ( text2, this, slot, 0, id );
#define add_separator_item(id,text2,slot) \
    if ( !menu->text ( id ).isNull () ) {\
        menu->removeItemAt ( menu->indexOf ( id ) - 1 );\
        menu->removeItem ( id );\
    }\
    menu->insertSeparator ();\
    menu->insertItem ( text2, this, slot, 0, id );
#define add_item_icon(id,icon,text2,slot) \
    if ( !menu->text ( id ).isNull () && menu->indexOf ( id ) != -1 ) {\
        menu->removeItem ( id );\
    }\
    menu->insertItem ( SmallIcon(icon), text2, this, slot, 0, id );
#define add_separator_item_icon(id,icon,text2,slot) \
    if ( !menu->text ( id ).isNull () ) {\
        menu->removeItemAt ( menu->indexOf ( id ) - 1 );\
        menu->removeItem ( id );\
    }\
    menu->insertSeparator ();\
    menu->insertItem ( SmallIcon(icon), text2, this, slot, 0, id );

TaskMenuManager::TaskMenuManager ( QPopupMenu *menu, Task *task ):
    QObject ( menu ) {
    this->task = task;
    this->menu = menu;
    add_item ( TMM_ACTIVATE_MENU, "Activate", SLOT(activate()) );
    add_item ( TMM_MINIMIZE_MENU, "Minimize", SLOT(minimize()) );
    add_item ( TMM_MAXIMIZE_MENU, "Maximize", SLOT(maximize()) );
    add_item ( TMM_RESTORE_MENU, "Restore", SLOT(restore()) );
    add_item ( TMM_SHADE_MENU, "Shade", SLOT(shade()) );
    add_item_icon ( TMM_ALWAYSONTOP_MENU, "attach", "Always on Top", SLOT(alwaysontop()) );
    desktops = new QPopupMenu ( menu );
    desktops->insertItem ( "All Desktops", 1 );
    desktops->insertSeparator ();
    for ( int i = 1; i <= MenuBarContainer::taskmanager->numberOfDesktops (); i++ )
      desktops->insertItem ( QString ().sprintf ( "Desktop %i", i ), i + 1 );
    connect ( desktops, SIGNAL(activated(int)), this, SLOT(desktop(int)) );
    if ( !menu->text ( TMM_DESKTOP_MENU ).isNull () ) {
        menu->removeItem ( TMM_DESKTOP_MENU );
    }
    menu->insertItem ( "To Desktop", desktops, TMM_DESKTOP_MENU );
    add_separator_item_icon ( TMM_KILL_MENU, "stop", "Kill", SLOT(kill()) );
    add_item_icon ( TMM_CLOSE_MENU, "remove", "Close", SLOT(close()) );
    if ( task ) {
        connect( task, SIGNAL(changed()), this, SLOT(changed()) );
        changed ();
    }
}

void TaskMenuManager::changed () {
    if ( task->isIconified() ) {
        menu->setItemEnabled ( TMM_MINIMIZE_MENU, false );
	menu->setItemEnabled ( TMM_RESTORE_MENU, true );
	menu->setItemEnabled ( TMM_MAXIMIZE_MENU, true );
    }
    else if ( task->isMaximized() ) {
        menu->setItemEnabled ( TMM_MINIMIZE_MENU, true );
	menu->setItemEnabled ( TMM_RESTORE_MENU, true );
	menu->setItemEnabled ( TMM_MAXIMIZE_MENU, false );
    }
    else {
        menu->setItemEnabled ( TMM_MINIMIZE_MENU, true );
	menu->setItemEnabled ( TMM_RESTORE_MENU, false );
	menu->setItemEnabled ( TMM_MAXIMIZE_MENU, true );
    }
    menu->setItemChecked ( TMM_SHADE_MENU, task->isShaded () );
    menu->setItemChecked ( TMM_ALWAYSONTOP_MENU, task->isAlwaysOnTop () );
    for ( int i = 0; i <= MenuBarContainer::taskmanager->numberOfDesktops (); i++ )
        desktops->setItemChecked ( i + 1, false );
    if ( task->isOnAllDesktops () )
        desktops->setItemChecked ( 1, true );
    else
        desktops->setItemChecked ( task->desktop () + 1, true );
}

void TaskMenuManager::activate () {
    task->activate ();
}

void TaskMenuManager::minimize () {
    task->iconify ();
}

void TaskMenuManager::maximize () {
    task->maximize ();
}

void TaskMenuManager::restore () {
    task->restore ();
}

void TaskMenuManager::shade () {
    task->toggleShaded ();
}

void TaskMenuManager::alwaysontop () {
    task->toggleAlwaysOnTop ();
}

void TaskMenuManager::desktop ( int d ) {
    task->toDesktop ( d - 1 );
}

void TaskMenuManager::kill () {
    XKillClient ( qt_xdisplay (), task->window () );
    remove ( task->window () );
}

void TaskMenuManager::close () {
    task->close ();
    remove ( task->window () );
}

TasksMenuManager::TasksMenuManager ( QPopupMenu *menu, Task *task, int which ):
    QObject ( menu ) {
    this->task = task;
    this->menu = menu;
    this->which = which;
    add_item ( TMM_MINIMIZE_MENU, "Minimize", SLOT(minimize()) );
    add_item ( TMM_MAXIMIZE_MENU, "Maximize", SLOT(maximize()) );
    add_item ( TMM_RESTORE_MENU, "Restore", SLOT(restore()) );
    add_item ( TMM_SHADE_MENU, "Shade", SLOT(shade()) );
    desktops = new QPopupMenu ( menu );
    desktops->insertItem ( "All Desktops", 1 );
    desktops->insertSeparator ();
    for ( int i = 1; i <= MenuBarContainer::taskmanager->numberOfDesktops (); i++ )
      desktops->insertItem ( QString ().sprintf ( "Desktop %i", i ), i + 1 );
    connect ( desktops, SIGNAL(activated(int)), this, SLOT(desktop(int)) );
    if ( !menu->text ( TMM_DESKTOP_MENU ).isNull () ) {
        if ( menu->indexOf ( TMM_DESKTOP_MENU ) != -1 )
            menu->removeItem ( TMM_DESKTOP_MENU );
    }
    menu->insertItem ( "To Desktop", desktops, TMM_DESKTOP_MENU );
    add_item_icon ( TMM_CLOSE_MENU, "remove", "Close", SLOT(close()) );
}

void TasksMenuManager::performaction ( int action, int extraparam ) {
    QPtrListIterator<Task> it ( ctasklist );
    bool add;
    for ( ; it.current (); ++it ) {
        if ( it.current ()->desktop () == MenuBarContainer::taskmanager->currentDesktop () || it.current ()->isOnAllDesktops () )
	    switch ( which ) {
	        case OtherThanTaskGroup: add = !Task::idMatch ( it.current ()->classClass (), task->classClass () ); break;
	        case OtherThanTask: add = ( it.current () != task ); break;
	        case AllOfTaskGroup: add = Task::idMatch ( it.current ()->classClass (), task->classClass () ); break;
		default: add = false; // just to make sure a default case
	    }
        else
	    add = false;
        if ( add ) {
	    switch ( action ) {
	        case TMM_CLOSE_MENU: it.current ()->close (); break;
		case TMM_DESKTOP_MENU: it.current ()->toDesktop ( extraparam ); break;
		case TMM_MINIMIZE_MENU: it.current ()->iconify (); break;
		case TMM_MAXIMIZE_MENU: it.current ()->maximize (); break;
		case TMM_RESTORE_MENU: it.current ()->restore (); break;
		case TMM_SHADE_MENU: it.current ()->toggleShaded (); break;
	    }
	}
    }
}

void TasksMenuManager::minimize () {
    performaction ( TMM_MINIMIZE_MENU, 0 );
}

void TasksMenuManager::maximize () {
    performaction ( TMM_MAXIMIZE_MENU, 0 );
}

void TasksMenuManager::restore () {
    performaction ( TMM_RESTORE_MENU, 0 );
}

void TasksMenuManager::close () {
    performaction ( TMM_CLOSE_MENU, 0 );
}

void TasksMenuManager::shade () {
    performaction ( TMM_SHADE_MENU, 0 );
}

void TasksMenuManager::desktop ( int d ) {
    performaction ( TMM_DESKTOP_MENU, d - 1 );
}

