/*****************************************************************

Copyright (c) 2001 John Firebaugh <jfirebaugh@kde.org>
Copyright (c) 2002 Siegfried Nijssen <snijssen@liacs.nl>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#ifndef __menubarcontainer_h__
#define __menubarcontainer_h__

#include <kpanelapplet.h>
#include <kpanelextension.h>
#include <dcopobject.h>

#include <fittslawframe.h>

class QBoxLayout;
class QWidgetStack;
class QLabel;
class QTimer;
class QMenuBar;
class QPopupMenu;
class QPushButton;

class MyTaskManager;
class Task;
class MenuBarContainer;
class KAboutKDE;

class Task;

class TaskMenuManager:public QObject {
    Q_OBJECT
public:
    TaskMenuManager ( QPopupMenu *menu, Task *task );
private:
    QPopupMenu *menu, *desktops;
    Task *task;
public slots:
    void minimize ();
    void maximize ();
    void restore ();
    void kill ();
    void close ();
    void changed ();
    void activate ();
    void shade ();
    void alwaysontop ();
    void desktop(int);
signals:
    void remove(int);
};


class TasksMenuManager:public QObject {
    Q_OBJECT
public:
    enum { OtherThanTaskGroup = 1, AllOfTaskGroup = 2, OtherThanTask = 3};
    TasksMenuManager ( QPopupMenu *menu, Task *task, int which );
private:
    QPopupMenu *menu, *desktops;
    Task *task;
    int which;
    void performaction ( int action, int extraparam );
public slots:
    void minimize ();
    void maximize ();
    void restore ();
    void close ();
    void desktop(int);
    void shade ();
};

class MenuBarContainer : public FittsLawFrame, virtual public DCOPObject
{
    Q_OBJECT
    K_DCOP
    friend class TaskMenuManager;
    friend class TasksMenuManager;
public:
    MenuBarContainer( bool enableFrame, QWidget* parent = 0, const char* name = 0 );
    ~MenuBarContainer();

    void orientationChange( Orientation );

    void preferences();
    // FIXME: need a globally defined enum Position
    QSize sizeHint( KPanelExtension::Position, QSize maxSize ) const;

    int widthForHeight ( int height );
k_dcop:
    virtual WId addMenubar ( WId /* the menubar which is added */, WId /* the window to which the menubar belongs */ );
    virtual void setNoWindowMenubar ( WId );

protected slots:
    void configure();
    void removeMenubar ( Task* );
    void raiseDefaultMenubar();
    void activeWindowChanged(WId);
    void newWindow ();
    void helpCurrentWindow();
    void helpAboutKDE();
    void removeTaskMenu(int);
    void copy();
    void paste();
    void slotChangeStyle ();

protected:

private:
    KPanelApplet::Direction      direction;
    QBoxLayout *                 layout;
    QWidgetStack *		 menubars;
    QMenuBar *			 defaultMenubar;
    WId			 	 noWindowMenubar;
    QPopupMenu *helpmenu, *editmenu, *windowmenu, *allwindowmenu, *otherwindowsmenu;
    void createAllWinMenu ( const QString &name );
    void MenuBarContainer::sendkey ( long mask, long key );
    QLabel *applabel;
    static MyTaskManager *	 taskmanager;
    Task *currenttask;
    TaskMenuManager *windowmenumanager;
    TasksMenuManager *otherwindowsmenumanager;

    MyTaskManager *taskManager ();
    KAboutKDE *aboutkde;
};

#endif
